@@ -1,4 +1,5 @@
 # -*- coding: UTF-8 -*-
+import hashlib
 import time
 import logging
 import os
@@ -12,6 +13,7 @@ try:
         NoSuchElementException, ElementNotInteractableException,
         InvalidElementStateException, WebDriverException,
     )
+    from selenium.webdriver.common.by import By
     from selenium.webdriver.common.keys import Keys
     from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
     from selenium.webdriver.support.ui import WebDriverWait, Select
@@ -29,11 +31,14 @@ logger = logging.getLogger('AUTOSCRAPE')
 
 
 class SeleniumBrowser(BrowserBase, Tagger):
+    # override a None passed to page_timeout because we need
+    # to set an integer value
+    DEFAULT_TIMEOUT = 30
 
     def __init__(self, driver="Firefox", leave_host=False,
                  load_images=False, form_submit_natural_click=False,
                  form_submit_wait=5, output=None, show_browser=False,
-                 browser_binary=None,
+                 browser_binary=None, page_timeout=None,
                  remote_hub="http://localhost:4444/wd/hub", **kwargs):
         try:
             webdriver
@@ -45,6 +50,11 @@ class SeleniumBrowser(BrowserBase, Tagger):
             )
             sys.exit(1)
 
+        # toggle the various timeouts in selenium
+        self.timeout = page_timeout
+        if self.timeout is None:
+            self.timeout = self.DEFAULT_TIMEOUT
+
         # Needs geckodriver:
         # https://github.com/mozilla/geckodriver/releases
         # Version 0.20.1 is recommended as of 14/07/2018
@@ -81,7 +91,7 @@ class SeleniumBrowser(BrowserBase, Tagger):
                 firefox_profile=firefox_profile,
                 firefox_binary=binary,
             )
-            self.driver.set_page_load_timeout(30)
+            self.driver.set_page_load_timeout(self.timeout)
 
         # this requires chromedriver to be on the PATH
         # if using chromium and ubuntu, apt install chromium-chromedriver
@@ -130,8 +140,12 @@ class SeleniumBrowser(BrowserBase, Tagger):
         # this is in the form of (command, *args, **kwargs)
         self.path = []
         # store the values of the history stack length (via JS) for use in determining
-        # whether or not we actually navigated somewhere
+        # whether or not we actually navigated somewhere. stored in window, hist
+        # pairs: [[win_1, 1], ..., [win_N, N]]
         self.history_stack = []
+        # used for detecting infinite loop inside a form (if the 'next' button
+        # never does disabled once at the end of results)
+        self.page_hashes = []
         # tree building
         self.graph = Graph()
         # sometimes the firefox driver loses its pipe to the browser. in
@@ -189,14 +203,27 @@ class SeleniumBrowser(BrowserBase, Tagger):
         script = "return window.history.length;"
         return self._driver_exec(self.driver.execute_script, script)
 
+    def _get_open_windows(self):
+        """
+        Get the handles of each open browser window and add any new
+        ones to the open window stack.
+        """
+        return self.driver.window_handles
+
+    def _switch_to_window(self, window_id):
+        self.driver.switch_to.window(window_id)
+
+    def _switch_to_iframe(self, frame_id):
+        self.driver.switch_to.frame(frame_id)
+
     def _wait_check(self, driver):
         """
         This is the check that gets ran to determine whether
         the page is loaded or not.
         """
         logger.debug(" - Waiting for page to load (document.readyState)...")
-        script = "return document.readyState;"
-        result = self._driver_exec(self.driver.execute_script, script)
+        script = "return document.readyState"
+        result = driver.execute_script(script)
         return result == "complete"
 
     def _loadwait(self, fn, *args, **kwargs):
@@ -204,7 +231,6 @@ class SeleniumBrowser(BrowserBase, Tagger):
         Run a driver interaction function, wait for the page to
         become ready, and handle any broken pipe errors
         """
-        wait_for_stale_time = 10  # seconds
         start = time.time()
         check_alerts = False
         if "check_alerts" in kwargs:
@@ -232,33 +258,9 @@ class SeleniumBrowser(BrowserBase, Tagger):
 
         # wait for the page to become ready, up to 30s, checks every 0.5s
         logger.debug(" - Performing native WebDriverWait...")
-        wait = WebDriverWait(self.driver, 30)
+        wait = WebDriverWait(self.driver, self.timeout)
         wait.until(self._wait_check)
 
-        stale_check_max_times = 30.0
-        stale_check_times = 0
-        while stale_check_times < stale_check_max_times:
-            logger.debug(" - Doing ID check (no. %s)..." % (stale_check_times))
-            if elem.id != self.driver.find_element_by_xpath("html").id:
-                break
-            stale_check_times += 1
-            time.sleep(wait_for_stale_time / stale_check_max_times)
-
-        stale_check_max_times = 30.0
-        stale_check_times = 0
-        while stale_check_times < stale_check_max_times:
-            logger.debug(" - Doing stale check (no. %s)..." % (stale_check_times))
-            try:
-                elem.text
-            except StaleElementReferenceException:
-                logger.debug("[.] Stale element found! Loading complete.")
-                break
-            except NoSuchElementException:
-                logger.debug("[.] Stale element found! Loading complete.")
-                break
-            stale_check_times += 1
-            time.sleep(wait_for_stale_time / stale_check_max_times)
-
         t = time.time() - start
         logger.debug(" - Wait for load succeeded in %s" % t)
 
@@ -281,7 +283,11 @@ class SeleniumBrowser(BrowserBase, Tagger):
         logger.info("[.] Fetching %s" % url)
         self._loadwait(self.driver.get, url)
         self.path.append(("fetch", (url,), {}))
-        self.history_stack.append(self._get_history_depth())
+
+        depth = self._get_history_depth()
+        window = self._get_open_windows()[-1]
+        self.history_stack.append([window, depth])
+
         node = "Fetch\n url: %s" % url
         self.graph.add_root_node(node, url=url, action="fetch")
 
@@ -290,15 +296,28 @@ class SeleniumBrowser(BrowserBase, Tagger):
         logger.debug(" - current path-length=%s path=%s" % (
             len(self.path), self._no_tags(self.path),
         ))
-        logger.debug("History stack: %s" % (self.history_stack))
+
         # only go 'back' if the history depth changed from
         # now to the previous (back) value. if they're the
         # same, we just clicked something inside the page
         if len(self.history_stack) > 1:
-            current_hist = self.history_stack.pop()
-            if current_hist != self.history_stack[-1]:
-                logger.debug("History depth changed since last action.")
+            window, depth = self.history_stack.pop()
+            prev_win, prev_depth = self.history_stack[-1]
+            # same window, we just incremented the history stack once
+            if window == prev_win and depth != prev_depth:
+                logger.debug(" - History depth changed since last action.")
+                # going backward should be the only time we have a
+                # frame focused, where we want to be on the main page
+                self.driver.switch_to.default_content()
                 self._loadwait(self.driver.back)
+            # time to close an open window, focus old
+            elif window != prev_win:
+                logger.debug(" - Going back to previous window")
+                # close current window (make sure we've focused proper one)
+                self._switch_to_window(window)
+                self.driver.close()
+                # focus previous one
+                self._switch_to_window(prev_win)
         self.path.pop()
         self.graph.move_to_parent()
 
@@ -314,41 +333,6 @@ class SeleniumBrowser(BrowserBase, Tagger):
         script = "arguments[0].target='_self';"
         self._driver_exec(self.driver.execute_script, script, elem)
 
-    def element_by_tag(self, tag):
-        """
-        Take a tag and return the corresponding live element in the DOM.
-        """
-        inside_id = False
-        # escaping logic
-        newtag = ""
-        for c in tag:
-            if c == "#":
-                inside_id = True
-                newtag += c
-                continue
-
-            # end of ID
-            elif inside_id and re.search("\s", c):
-                inside_id = False
-
-            elif inside_id and c in self.css_escapables:
-                for escapable in self.css_escapables:
-                    c = "\%s" % escapable
-
-            newtag += c
-
-        if newtag != tag:
-            logger.debug("Original tag: %s, newtag: %s" % (tag, newtag))
-            tag = newtag
-
-        try:
-            return self._driver_exec(
-                self.driver.find_element_by_css_selector, tag
-            )
-        except Exception as e:
-            msg = "Error finding element for tag %s. Error: %s"
-            logger.error(msg % (tag, e))
-
     def click(self, tag, iterating_form=False):
         """
         Click an element by a given tag. Returns True if the link
@@ -359,12 +343,25 @@ class SeleniumBrowser(BrowserBase, Tagger):
             logger.warn("Element by tag not found. Tag: %s" % tag)
             return False
 
-        name = self._driver_exec(elem.tag_name)
-        onclick = self._driver_exec(elem.get_attribute, "onclick")
+        try:
+            wait = WebDriverWait(self.driver, self.timeout)
+            wait.until(EC.element_to_be_clickable((By.CSS_SELECTOR, tag)))
+        except TimeoutException as e:
+            pass
+
+        hash_parts = []
+        hash_parts.append(self._driver_exec(elem.tag_name))
         href = self._driver_exec(elem.get_attribute, "href")
-        hash = "%s|%s|%s" % (href, onclick, name)
+        hash_parts.append(href)
+        hash_parts.append(self._driver_exec(elem.get_attribute, "onclick"))
         text = self._driver_exec(elem.text)
+        hash = "|".join([h for h in hash_parts if h])
+        logger.debug(" - Link hash : %s" % (hash))
+
+        # only enable this for URL links
+        # if href and hash in self.visited and not iterating_form:
         if hash in self.visited and not iterating_form:
+            logger.debug("Link already clicked! Hash: %s" % (hash))
             return False
 
         self.visited.add(hash)
@@ -389,13 +386,29 @@ class SeleniumBrowser(BrowserBase, Tagger):
                 logger.error("[!] Current URL: %s" % (self.page_url))
                 return False
         except Exception as e:
-            logger.error("[!] Error clicking %s: %s" % (href, e))
-            logger.error("[!] Stacktrace: %s" % (e.stacktrace))
-            logger.error("[!] Screen: %s" % (e.screen))
-            logger.error("[!] Current URL: %s" % (self.page_url))
+            logger.debug("[!] Error clicking: %s" % (e))
+            logger.debug("[!] Current URL: %s" % (self.page_url))
             return False
 
-        self.history_stack.append(self._get_history_depth())
+        # apply form submit waits to 'next' button clicks
+        if iterating_form and self.form_submit_wait:
+            logger.debug(
+                " - Forcing post-submit wait period of %ss" %
+                self.form_submit_wait
+            )
+            time.sleep(self.form_submit_wait)
+
+        depth = self._get_history_depth()
+        window = self._get_open_windows()[-1]
+
+        if len(self.history_stack) > 1:
+            prev_window = self.history_stack[-1][0]
+            if window != prev_window:
+                logger.debug(" - Focusing new window: %s" % (window))
+                self._switch_to_window(window)
+
+        self.history_stack.append([window, depth])
+
         self.path.append((
             "click", (tag,), {"iterating_form": iterating_form}
         ))
@@ -428,7 +441,7 @@ class SeleniumBrowser(BrowserBase, Tagger):
             "(\[:left:\]|\[:right:\]|\[:down:\]|\[:up:\]|\[:enter:\])",
             input
         )
-        logger.debug("Split replacements input string: %s" % split_replacements)
+        logger.debug(" - Split replacements input string: %s" % split_replacements)
         for chunk in split_replacements:
             if not chunk:
                 continue
@@ -441,7 +454,7 @@ class SeleniumBrowser(BrowserBase, Tagger):
         """
         Enter some input into an element by a given tag.
         """
-        logger.info("Injecting text \"%s\" to input" % (input))
+        logger.info(" - Injecting text \"%s\" to input" % (input))
         elem = self.element_by_tag(tag)
         self._driver_exec(self.scrolltoview, elem)
         try:
@@ -465,7 +478,7 @@ class SeleniumBrowser(BrowserBase, Tagger):
         """
         Select an input select option based on its visible string value.
         """
-        logger.info("Selecting option %s" % (option_str))
+        logger.info(" - Selecting option %s" % (option_str))
         elem = self.element_by_tag(tag)
         self._driver_exec(self.scrolltoview, elem)
         # TODO: wrap in _driver_exec after testing
@@ -480,18 +493,19 @@ class SeleniumBrowser(BrowserBase, Tagger):
         }
         self.graph.add_action_to_current(action)
 
-    def input_checkbox(self, tag, to_check):
+    def input_checkbox(self, tag, to_check, radio=False):
         """
         Check, uncheck, or don't touch an input checkbox, based
         on its current checked value.
         """
-        logger.info("Checking to checkbox selected=%s" % (to_check))
+        logger.info(" - Checking to checkbox selected=%s" % (to_check))
         elem = self.element_by_tag(tag)
         self._driver_exec(self.scrolltoview, elem)
         # TODO: wrap in _driver_exec after testing
         if elem.is_selected() != to_check:
             elem.click()
-            self._driver_exec(elem.clear)
+            if not radio:
+                self._driver_exec(elem.clear)
         self.path.append(("input_checkbox", (tag, to_check,), {}))
         action = {
             "action": "input_checkbox",
@@ -505,31 +519,25 @@ class SeleniumBrowser(BrowserBase, Tagger):
         Perform a "natural" click of an element, so that if there are
         any parent listeners, they will also be triggered.
         """
-        logger.info("Performing a 'natural' click")
+        logger.info(" - Performing a 'natural' click")
         from selenium.webdriver.common.action_chains import ActionChains
         ac = ActionChains(self.driver)
         x_off = 0
         y_off = 0
         ac.move_to_element(elem).move_by_offset(x_off, y_off).click().perform()
 
-    def submit(self, tag):
-        """
-        Submit a form from a given tag. Assumes all inputs are filled.
-        """
-        logger.info("Submitting form by tag: %s" % tag)
-        form = self.element_by_tag(tag)
-        self._driver_exec(self.scrolltoview, form)
-
-        # try to find a Submit input button
+    def find_submit_button(self, form):
         sub = None
-        try:
-            sub = self._driver_exec(
-                form.find_element_by_xpath,
-                "//input[@type='submit']"
-            )
-            logger.debug(" - Form submit input button: %s" % (sub))
-        except NoSuchElementException as e:
-            pass
+        submit_btns = self._driver_exec(
+            form.find_elements_by_xpath,
+            "//input[@type='submit']"
+        )
+        for el in submit_btns:
+            txt = el.get_attribute("value").lower()
+            if "submit" in txt or "search" in txt or txt == "go":
+                sub = el
+                logger.debug(" - Form submit input button: %s" % (txt))
+                break
 
         # try to find a Submit link
         # NOTE: instead of using xpath we're going to do a manual search.
@@ -549,6 +557,20 @@ class SeleniumBrowser(BrowserBase, Tagger):
             except NoSuchElementException as e:
                 pass
 
+        return sub
+
+    def submit(self, tag):
+        """
+        Submit a form from a given tag. Assumes all inputs are filled.
+        """
+        logger.info("[.] Submitting form.")
+        logger.debug(" - form button tag: %s" % tag)
+        form = self.element_by_tag(tag)
+        self._driver_exec(self.scrolltoview, form)
+
+        # try to find a Submit input button
+        sub = self.find_submit_button(form)
+
         # attempt to click on the button, if this fails, we will
         # try to perform a form submit
         sub_failure = False
@@ -567,7 +589,11 @@ class SeleniumBrowser(BrowserBase, Tagger):
             logger.debug("Using form.submit selenium shim")
             self._loadwait(form.submit, check_alerts=True)
 
-        self.history_stack.append(self._get_history_depth())
+        depth = self._get_history_depth()
+        window = self._get_open_windows()[-1]
+        # assuming we can't pop a new window here, for now
+        self.history_stack.append([window, depth])
+
         self.path.append(("submit", (tag,), {}))
         node = "Submit\n tag: %s" % (tag)
         node_meta = {
@@ -579,7 +605,7 @@ class SeleniumBrowser(BrowserBase, Tagger):
         # TODO: better way to wait for this, post-alert clicked
         if self.form_submit_wait:
             logger.debug(
-                "Forcing post-submit wait period of %ss" %
+                " - Forcing post-submit wait period of %ss" %
                 self.form_submit_wait
             )
             time.sleep(self.form_submit_wait)
@@ -598,6 +624,40 @@ class SeleniumBrowser(BrowserBase, Tagger):
     def page_url(self):
         return self._driver_exec(self.driver.current_url)
 
+    @property
+    def infinite_loop_detected(self):
+        """
+        Store the recent page hashes and check to see if
+        the most recent pages have all been identical. This
+        determines an infinite loop. Note that this is only
+        intended to be used inside a results pages loop.
+        """
+        infinite_loop_threshold = 5
+
+        html_b = self.page_html
+        if type(html_b) != bytes:
+            html_b = self.page_html.encode("utf-8")
+
+        hash = hashlib.md5(html_b).hexdigest()
+        logger.debug(" - Page hash: %s" % (hash))
+
+        # don't even look for infinite loop until we have the
+        # required number of hashes to search
+        if len(self.page_hashes) < infinite_loop_threshold:
+            self.page_hashes.append(hash)
+            return False
+
+        self.page_hashes = self.page_hashes[1:] + [hash]
+        recent = self.page_hashes[:infinite_loop_threshold]
+        identical = recent.count(hash)
+        logger.debug(" - %s pages of the past %s have been identical" % (
+            identical, infinite_loop_threshold
+        ))
+        if identical == infinite_loop_threshold:
+            return True
+
+        return False
+
     def get_clickable(self, type=None):
         """
         Get tags, by type (optional), for the currently loaded page.
