@@ -38,13 +38,14 @@ class Controller:
                  remote_hub="http://localhost:4444/wd/hub", output=None,
                  form_submit_natural_click=False, form_submit_wait=5,
                  warc_index_file=None, warc_directory=None,
-                 load_images=False, show_browser=False,
+                 load_images=False, show_browser=False, page_timeout=None,
                  html_embeddings_file=None, word_embeddings_file=None,
                  backend="selenium", vectorizer="text"):
         """
         Set up our WebDriver and misc utilities.
         """
         Browser = None
+        self.backend = backend
         if backend == "selenium":
             Browser = SeleniumBrowser
         elif backend == "requests":
@@ -63,7 +64,7 @@ class Controller:
             form_submit_wait=form_submit_wait,
             warc_index_file=warc_index_file, warc_directory=warc_directory,
             load_images=load_images, show_browser=show_browser,
-            output=output,
+            output=output, page_timeout=page_timeout,
         )
 
         Vectorizer = None
@@ -96,6 +97,7 @@ class Controller:
         #   other forms ...,
         # ]
         self.inputs = []
+        self.force_wait = 5
 
     def load_indices(self):
         logger.debug("[.] Loading page vectors...")
@@ -154,17 +156,21 @@ class Controller:
         gets the links for the current page and sets its tag array.
         """
         self.scraper.fetch(url, initial=True)
+        if self.backend == "selenium":
+            time.sleep(self.force_wait)
         self.load_indices()
 
-    def select_link(self, index):
+    def select_link(self, index, iterating_form=False):
         if index >= len(self.clickable):
             logger.error(
                 "[!] Critical error: link index exceeds clickable length."
             )
             return False
         tag = self.clickable[index]
-        clicked = self.scraper.click(tag)
+        clicked = self.scraper.click(tag, iterating_form=iterating_form)
         if clicked:
+            if self.backend == "selenium":
+                time.sleep(self.force_wait)
             self.load_indices()
         return clicked
 
@@ -172,6 +178,8 @@ class Controller:
         tag = self.buttons[index]
         clicked = self.scraper.click(tag, iterating_form=iterating_form)
         if clicked:
+            if self.backend == "selenium":
+                time.sleep(self.force_wait)
             self.load_indices()
         return clicked
 
@@ -204,11 +212,24 @@ class Controller:
         tag = self.inputs[form_ix][3][index]
         self.scraper.input(tag, chars)
 
+    def input_radio_option(self, form_ix, index, radio_index):
+        """
+        Select a radio checkbox from a given form, checkbox group
+        index and desired value by text.
+        """
+        grp_tags = self.inputs[form_ix][4][index]
+        tag = grp_tags[radio_index]
+        self.scraper.input_checkbox(tag, True, radio=True)
+
     def submit(self, index):
         tag = self.forms[index]
         self.scraper.submit(tag)
+        if self.backend == "selenium":
+            time.sleep(self.force_wait)
         self.load_indices()
 
     def back(self):
         self.scraper.back()
+        if self.backend == "selenium":
+            time.sleep(self.force_wait)
         self.load_indices()
